# Author: Mark Blakeney, Feb 2024.
'''
Run an installed application using a debugger.

Tries to work out your preferred debugger from the standard
PYTHONBREAKPOINT environment variable. If not set it defaults to pdb. Or
you can set it explicitly with the -d/--debugger option.
'''

from __future__ import annotations

import os
import re
from argparse import ArgumentParser, Namespace
from typing import Optional

from .. import utils
from ..run import run

def init(parser: ArgumentParser) -> None:
    'Called to add command arguments to parser at init'
    parser.add_argument('-e', '--executable',
                        help='executable to run, '
                        'default is same as "package" name')
    parser.add_argument('-d', '--debugger',
                        help='explicit debugger package to use')
    parser.add_argument('package',
                        help='installed application name')
    parser.add_argument('args', nargs='*',
                        help='options and arguments to pass to application, '
                        'should start with "--"')

def main(args: Namespace) -> Optional[str]:
    'Called to action this command'
    pkgname, vdir = utils.get_package_from_arg(args.package, args)
    if not vdir:
        return f'Application {pkgname} is not installed.'

    if args.debugger:
        debugger = args.debugger
    else:
        debugger = re.sub(r'\..*', '', os.getenv('PYTHONBREAKPOINT', ''))
        if not debugger or debugger == '0':
            debugger = 'pdb'

    python = str(utils.vdir_bin(vdir) / 'python')
    exe = str(args._bin_dir / (args.executable or pkgname))
    run([python, '-m', debugger, exe] + args.args, quiet=True)
    return None



# Author: Mark Blakeney, Feb 2024.
'Uninstall extra packages from an application.'
from __future__ import annotations

from argparse import ArgumentParser, Namespace
from typing import Optional

from .. import utils

def init(parser: ArgumentParser) -> None:
    'Called to add command arguments to parser at init'
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='give more output')
    parser.add_argument('package',
                        help='installed application name')
    parser.add_argument('extras', nargs='+',
                        help='extra package name[s] to uninstall')

def main(args: Namespace) -> Optional[str]:
    'Called to action this command'
    pkgname, vdir = utils.get_package_from_arg(args.package, args)
    if not vdir:
        return f'Application {pkgname} is not installed.'

    pip_args = ['uninstall'] + \
            utils.make_args((args.verbose, '-v')) + args.extras
    if not utils.piprun(vdir, args, pip_args):
        return f'Error: failed to uninstall "{args.extras}" from {pkgname}'

    return utils.add_or_remove_pkg(vdir, args, pkgname, args.extras, add=False)



# Author: Mark Blakeney, Feb 2024.
'Install one or more Python applications using isolated virtual environments.'
from __future__ import annotations

from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Optional

from filelock import FileLock

from .. import utils
from ..run import run

MAX_VDIRS = 1_000_000

def _get_next_vdir(vdirbase: Path) -> Optional[Path]:
    'Return the first available venv directory'
    vdirs = set(int(f.name) for f in vdirbase.iterdir())
    for n in range(1, MAX_VDIRS + 1):
        if n not in vdirs:
            return vdirbase / str(n)

    return None

def init(parser: ArgumentParser) -> None:
    'Called to add command arguments to parser at init'
    parser.add_argument('-p', '--python',
                        help='specify explicit python executable path')
    parser.add_argument('-f', '--force', action='store_true',
                        help='recreate any already installed venv')
    parser.add_argument('-e', '--editable', action='store_true',
                        help='install application[s] in editable mode')
    parser.add_argument('-d', '--include-deps', action='store_true',
                        help='include executables from dependencies')
    parser.add_argument('--system-site-packages', action='store_true',
                        help='allow venv access to system packages')
    parser.add_argument('-i', '--index-url',
                        help='base URL of Python Package Index')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='give more output')
    parser.add_argument('package', nargs='+',
                        help='application[s] to install')

def main(args: Namespace) -> Optional[str]:
    'Called to action this command'
    pyexe = str(utils.get_python(args))
    venv_args = [args._uv, 'venv', '-p', pyexe] + utils.make_args(
            (args.verbose, '-v'), (not args.verbose, '-q'),
            (args.system_site_packages, '--system-site-packages'))

    pip_args = 'install --compile'.split() + utils.make_args(
            (args.verbose, '-v'), (args.index_url, '-i', args.index_url))
    pip_earg = utils.make_args((args.editable, '-e'))

    vdirbase = args._venvs_dir
    for pkg in args.package:
        # Use a lock file in case we are running multiple installs in parallel
        with FileLock(args._lockfile):
            vdir = _get_next_vdir(vdirbase)
            if not vdir:
                return f'Error: Too many vdirs (>{MAX_VDIRS}) in {vdirbase}'

            # Create the vdir
            if not run(venv_args + [str(vdir)]):
                utils.rm_vdir(vdir, args)
                return f'Error: failed to create {vdir} for {pkg}.'

        python_exe = str((utils.vdir_bin(vdir) / 'python').resolve())
        python_ver = run((python_exe, '-V'), capture=True, ignore_error=True)
        python_ver = python_ver.strip().split()[1] if python_ver else '?ver?'
        print(f'Created "{vdir}" using "{python_exe}" ({python_ver})')

        # Install the package
        if not utils.piprun(vdir, args,
                            pip_args + ['--no-deps'] + pip_earg + [pkg]):
            utils.rm_vdir(vdir, args)
            return f'Error: failed to preinstall "{pkg}".'

        if not (versions := utils.get_versions(vdir, args)):
            utils.rm_vdir(vdir, args)
            return f'Error: failed to get versions for {pkg}.'

        if len(versions) != 1:
            return f'Error: multiple packages qualified: {list(versions)}'

        pkgname, (vers, editpath) = versions.popitem()
        pdir = Path(args._packages_dir, pkgname)

        if pdir.exists():
            if not args.force:
                utils.rm_vdir(vdir, args)
                return f'Error: venv for {pkgname} exists. Use -f to force.'
            print(f'Removing pre-existing {pkgname} venv dir.')
            utils.rm_vdir(pdir, args)
            pdir.unlink()

        if not utils.piprun(vdir, args, pip_args + pip_earg + [pkg]):
            utils.rm_vdir(vdir, args)
            return f'Error: failed to install "{pkg}".'

        pdir.symlink_to(vdir)

        data: dict = {'name': pkgname}
        if editpath:
            data['editpath'] = editpath

        if args.include_deps:
            data['deps'] = True

        if args.system_site_packages:
            data['sys'] = True

        if args.index_url:
            data['url'] = args.index_url

        if args.python:
            data['python'] = args.python

        if err := utils.make_links(vdir, pkgname, args, data):
            pdir.unlink()
            utils.rm_vdir(vdir, args)
            return err

    return None



# Author: Mark Blakeney, Feb 2024.
'Uninstall one, or more, or all applications.'
from __future__ import annotations

from argparse import ArgumentParser, Namespace
from typing import Optional

from .. import utils

def _uninstall(args: Namespace, pkgname: str) -> Optional[str]:
    'Uninstall given package'
    pkgname, vdir = utils.get_package_from_arg(pkgname, args)
    if not vdir:
        return f'Application {pkgname} is not installed.'

    if not utils.rm_package(pkgname, args):
        return f'Failed to uninstall {pkgname}.'

    print(f'{pkgname} uninstalled.')
    return None

def init(parser: ArgumentParser) -> None:
    'Called to add command arguments to parser at init'
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='give more output')
    parser.add_argument('--all', action='store_true',
                        help='uninstall ALL applications')
    parser.add_argument('--skip', action='store_true',
                        help='skip the specified applications when '
                        'uninstalling all (only can be specified with --all)')
    parser.add_argument('package', nargs='*',
                        help='application[s] to uninstall (or to skip for '
                        '--all --skip)')

def main(args: Namespace) -> Optional[str]:
    'Called to action this command'
    for pkgname in utils.get_package_names(args):
        if error := _uninstall(args, pkgname):
            return error

    return None



# Author: Mark Blakeney, Feb 2024.
'Run pip with given arguments on virtual environment for the given application.'
from __future__ import annotations

from argparse import ArgumentParser, Namespace
from typing import Optional

from .. import utils

def init(parser: ArgumentParser) -> None:
    'Called to add command arguments to parser at init'
    parser.add_argument('package',
                        help='installed application name')
    parser.add_argument('args', nargs='*',
                        help='arguments to pass to uv pip, '
                        'should start with "--".')

def main(args: Namespace) -> Optional[str]:
    'Called to action this command'
    pkgname, vdir = utils.get_package_from_arg(args.package, args)
    if not vdir:
        return f'Application {pkgname} is not installed.'

    if not utils.piprun(vdir, args, args.args, quiet=True):
        return f'Error: failed to run pip for {pkgname}'
    return None



# Author: Mark Blakeney, Feb 2024.
'''
Install extra packages into an application.

Note the same --index-url is used as/if specified in the original install.
'''
from __future__ import annotations

from argparse import ArgumentParser, Namespace
from typing import Optional

from .. import utils

def init(parser: ArgumentParser) -> None:
    'Called to add command arguments to parser at init'
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='give more output')
    parser.add_argument('package',
                        help='installed application name')
    parser.add_argument('extras', nargs='+',
                        help='extra package name[s] to inject/install')

def main(args: Namespace) -> Optional[str]:
    'Called to action this command'
    pkgname, vdir = utils.get_package_from_arg(args.package, args)
    if not vdir:
        return f'Application {pkgname} is not installed.'

    data = utils.get_json(vdir, args) or {}
    url = data.get('url')
    pip_args = 'install --compile'.split() + \
            utils.make_args((args.verbose, '-v'), (url, ('-i', url))) + \
            args.extras
    if not utils.piprun(vdir, args, pip_args):
        return f'Error: failed to install "{args.extras}" to {pkgname}'

    return utils.add_or_remove_pkg(vdir, args, pkgname, args.extras,
                                   data=data, add=True)



# Author: Mark Blakeney, Feb 2024.
'List installed application versions.'
from __future__ import annotations

from argparse import ArgumentParser, Namespace
from typing import Optional

from .. import utils

def init(parser: ArgumentParser) -> None:
    'Called to add command arguments to parser at init'
    parser.add_argument('package', nargs='?',
                        help='report specific application and dependent '
                        'package versions')

def main(args: Namespace) -> Optional[str]:
    'Called to action this command'
    def display(pkgname, version):
        ver, loc = version
        if loc:
            ver += f' @ {loc}'
        print(f'{pkgname}=={ver}')

    if args.package:
        pkgname, vdir = utils.get_package_from_arg(args.package, args)
        if not vdir:
            return f'Application {pkgname} not found.'

        if not (versions := utils.get_versions(args._packages_dir / pkgname,
                                               args)):
            return f'Application {pkgname} versions not found.'

        # Reorder version dict to put pkgname first
        if pkgname in versions:
            sversions = {pkgname: versions[pkgname]}
            del versions[pkgname]
            sversions.update(versions)
            versions = sversions

        for pkg, ver in versions.items():
            display(pkg, ver)

        return None

    for pdir, data in utils.get_all_pkg_venvs(args):
        package = pdir.name
        if versions := utils.get_versions(pdir, args):
            display(package, versions.get(package, ("unknown", None)))

    return None



# Author: Mark Blakeney, Feb 2024.
'Common module containing shared functions'
from __future__ import annotations

import json
import os
import platform
import shutil
import sys
from argparse import Namespace
from pathlib import Path
from typing import Iterable, Optional, Sequence

from .run import run

is_windows = platform.system() == 'Windows'

def subenvars(path: str) -> Path:
    'Substitute environment variables in a path string'
    return Path(os.path.expandvars(path)).expanduser()

def get_json(vdir: Path, args: Namespace) -> Optional[dict]:
    'Get JSON data for this virtual environment'
    tgt = vdir.resolve() / args._meta_file
    try:
        with tgt.open() as fp:
            return json.load(fp)
    except Exception:
        return None

def _set_json(vdir: Path, args: Namespace, data: dict) -> Optional[str]:
    'Set JSON data for this virtual environment'
    tgt = vdir.resolve() / args._meta_file
    try:
        with tgt.open('w') as fp:
            json.dump(data, fp)
    except Exception as e:
        return str(e)

    return None

def piprun(vdir: Path, args: Namespace, cmd: list[str],
           **kargs) -> Optional[str]:
    'Run given pip command in the virtual environment'
    # First element in cmd is the command, the rest are arguments. So
    # insert path to virtual environment after the command.
    cmd[1:1] = ['-p', str(vdir.resolve())]
    return run([args._uv, 'pip'] + cmd, **kargs)

def get_versions(vdir: Path, args: Namespace) -> \
        Optional[dict[str, tuple[str, Optional[str]]]]:
    'Return the versions of the packages in the virtual environment'
    if not (out := piprun(vdir, args, ['list'], capture=True)):
        return None

    found = False
    data: dict[str, tuple[str, Optional[str]]] = {}
    for line in out.splitlines():
        if line.startswith('-'):
            found = True
        elif found:
            fields = line.split(maxsplit=2)
            if len(fields) == 2:
                data[fields[0]] = fields[1], None
            else:
                data[fields[0]] = fields[1], fields[2]

    return data

def make_args(*args: Sequence) -> list[str]:
    'Build a list of args based on (bool, arg1, [arg2]) sequences'
    retlist: list[str] = []
    for alist in args:
        if alist[0]:
            retlist.extend(alist[1:])

    return retlist

def vdir_bin(vdir: Path) -> Path:
    'Return the bin directory for the virtual environment'
    return vdir / ('Scripts' if is_windows else 'bin')

def _load_record(rfile: Path, embedpath: str) -> Iterable[str]:
    'Yield the executable names from a RECORD file'
    with rfile.open() as fp:
        for line in fp:
            line = line.strip()
            if line.startswith(embedpath):
                yield Path(line.split(',', 1)[0]).name

def _link_app_files(vdir: Path, tgtdir: Path, pkgname: str,
                    args: Namespace, include_deps: bool) -> Iterable[str]:
    'Link app files from entry_points to tgtdir'
    vpath = vdir_bin(vdir)
    embedpath = '../../../bin/' if vpath.name == 'bin' \
            else f'..\\..\\{vpath.name}\\'

    pkgname = pkgname.replace('-', '_')
    for efile in vdir.glob('**/*.dist-info/RECORD'):
        if include_deps or efile.parent.name.startswith(f'{pkgname}-'):
            for app in _load_record(efile, embedpath):
                srcfile = vpath / app
                if srcfile.is_file() and not srcfile.is_symlink() and \
                      (srcfile.stat().st_mode & 0o111) == 0o111:
                    tgtfile = tgtdir / app
                    if tgtfile.is_symlink():
                        tgtfile.unlink()

                    if args.verbose:
                        print(f'Linking "{srcfile}" -> "{tgtfile}"')

                    tgtfile.parent.mkdir(parents=True, exist_ok=True)
                    tgtfile.symlink_to(srcfile)
                    yield srcfile.name

def _link_all_files(srcdir: Path, tgtdir: Path, pat: str,
                    verbose: bool) -> None:
    'Link files from srcdir to tgtdir'
    for srcfile in srcdir.glob(pat):
        tgtfile = tgtdir / srcfile.relative_to(srcdir)
        if tgtfile.is_symlink():
            tgtfile.unlink()

        if verbose:
            print(f'Linking "{srcfile}" -> "{tgtfile}"')

        tgtfile.parent.mkdir(parents=True, exist_ok=True)
        tgtfile.symlink_to(srcfile)

def _unlink_all_files(vdir: Path, args: Namespace) -> None:
    'Unlink all link files'
    dirlist: list[tuple[Path, str]] = [(args._bin_dir, '*')]
    if args._man_dir.is_dir():
        dirlist.append((args._man_dir, '*/*'))

    for srcdir, pat in dirlist:
        if srcdir.is_dir():
            for file in srcdir.glob(pat):
                if file.is_symlink() and vdir in file.resolve().parents:
                    if args.verbose:
                        print(f'Removing link "{file}"')
                    file.unlink()

def make_links(vdir: Path, pkgname: str, args: Namespace,
               data: Optional[dict]) -> Optional[str]:
    '[Re-]create app and man page links for package'

    vdir = vdir.resolve()

    # Unlink any existing links
    _unlink_all_files(vdir, args)

    # Link the package applications
    if not data and not (data := get_json(vdir, args)):
        return 'Error: No JSON data found.'

    include_deps = bool(args.include_deps if hasattr(args, 'include_deps')
                        else data.get('deps'))

    apps = list(_link_app_files(vdir, args._bin_dir, pkgname, args,
                                include_deps))

    # Link all the man pages
    if not args.no_man_pages:
        _link_all_files(vdir / 'share' / 'man', args._man_dir, '*/*',
                        args.verbose)

    # Save the apps in the JSON data
    if not apps:
        return f'Error: {pkgname} has no executables to install.'

    if not (freeze := piprun(vdir, args, ['freeze'], capture=True)):
        return 'Error: Failed to fetch freeze list.'

    (vdir / args._freeze_file).write_text(freeze)
    data['apps'] = sorted(apps)
    return _set_json(vdir, args, data)

def rm_vdir(vdir: Path, args: Namespace) -> None:
    'Remove all links that point into the virtual environment'
    vdir = vdir.resolve()
    _unlink_all_files(vdir, args)

    # Remove the venv
    if vdir.exists():
        if args.verbose:
            print(f'Removing "{vdir}"')

        shutil.rmtree(vdir)

def add_or_remove_pkg(vdir: Path, args: Namespace, pkgname: str,
                      pkgs: list[str], *,
                      add: bool, data: Optional[dict] = None) -> Optional[str]:
    'Record the addition/removal of a package into the virtual environment'
    if not data:
        data = get_json(vdir, args)
        if not data:
            return 'No JSON data found'

    dataset = set(data.get('injected') or [])
    pkgset = set(pkgs)
    newset = (dataset | pkgset) if add else (dataset - pkgset)

    if newset:
        data['injected'] = sorted(list(newset))
    else:
        data.pop('injected', None)

    return make_links(vdir, pkgname, args, data)

def rm_package(pkgname: str, args: Namespace) -> bool:
    'Remove this package'
    pdir = args._packages_dir / pkgname
    if not pdir.exists():
        return False

    vdir = pdir.resolve()

    if args.verbose:
        print(f'Removing link "{pdir}"')
    pdir.unlink()

    rm_vdir(vdir, args)
    return True

def get_all_pkg_venvs(args: Namespace) -> Iterable[tuple[Path, dict]]:
    'Return a list of all virtual environments and their JSON data'
    for pdir in sorted(args._packages_dir.iterdir()):
        if data := get_json(pdir, args):
            yield pdir, data

def _get_package_if_dir(name: str, args: Namespace) -> str:
    'Convert the given name if it corresponds to a package directory'
    if name not in {'.', '..'} and os.sep not in name:
        return name

    if not (namepath := Path(name).resolve()).is_dir():
        return name

    # Work out the package name from the current path. We look for
    # the closest parent (i.e. longest path) across all matching
    # application editpaths, unless we find an exact match then just
    # use that.
    candidates = {}
    for pdir, data in get_all_pkg_venvs(args):
        if data and (path := data.get('editpath')):
            # If we have an exact match then use it and ignore any
            # previous candidates.
            if (path := Path(path)) == namepath:
                return pdir.name

            if path in namepath.parents:
                # This path has a candidate parent, so record it.
                candidates[len(path.parts)] = pdir.name

    return candidates[max(candidates)] if candidates else name

def get_package_from_arg(name: str, args: Namespace) \
        -> tuple[str, Optional[Path]]:
    'Return the package name + vdir corresponding to the given arg, if any'
    name = _get_package_if_dir(name, args)
    path = (args._packages_dir / name).resolve()
    return name, (path if path.exists() else None)

def _rm_path(path: Path) -> None:
    'Remove the given path'
    print(f'Purging stray "{path}"', file=sys.stderr)
    if path.is_symlink():
        path.unlink()
    elif path.is_dir():
        shutil.rmtree(path)
    elif path.exists():
        path.unlink()

def purge_old_files(args: Namespace) -> None:
    'Clean out any old virtual environments, packages, and executables'
    # Remove any packages that do not point to a dir in the venvs directory
    valids_venvs = set()
    for pkg in args._packages_dir.iterdir():
        vdir = pkg.resolve()
        if vdir.parent != args._venvs_dir or not vdir.is_dir() \
                or not vdir.name.isdigit():
            _rm_path(pkg)
        else:
            valids_venvs.add(vdir.name)

    # Remove any venvs that are not in the packages directory
    for vdir in args._venvs_dir.iterdir():
        if vdir.name not in valids_venvs:
            _rm_path(vdir)

    # Remove any executables that point to a non-existent path
    for exe in args._bin_dir.iterdir():
        rexe = exe.resolve()
        if args._venvs_dir in rexe.parents and not rexe.exists():
            _rm_path(exe)

def get_package_names(args: Namespace) -> list[str]:
    'Return a list of validated package names based on command line args'
    if args.all:
        if not args.skip and args.package:
            args.parser.error('Can not specify package[s] with '
                            '--all unless also specifying --skip.')
    else:
        if args.skip:
            args.parser.error('--skip can only be specified with --all.')

        if not args.package:
            args.parser.error('Must specify at least one package, or --all.')

    given_names = set(_get_package_if_dir(p, args) for p in args.package)
    all_names = set(f.name for f in args._packages_dir.iterdir() if f.is_dir())

    if (unknown := given_names - all_names):
        s = 's' if len(unknown) > 1 else ''
        sys.exit(f'Error: Unknown package{s}: {", ".join(unknown)}')

    return sorted(all_names - given_names) if args.all else args.package

def get_python(args: Namespace) -> Path:
    'Return the python executable based on command line args'
    return subenvars(args.python) if args.python else args._pyexe

def version() -> str:
    'Return the version of this package'
    from importlib.metadata import version
    try:
        ver = version(Path(sys.argv[0]).stem)
    except Exception:
        ver = 'unknown'

    return ver



# Author: Mark Blakeney, Feb 2024.
'Reinstall one, or more, or all applications.'
from __future__ import annotations

import shutil
import tempfile
from argparse import ArgumentParser, Namespace
from copy import copy
from pathlib import Path
from typing import Optional

from .. import utils
from ..run import run

def _reinstall(args: Namespace, pkgname: str,
              venv_args: list[str]) -> Optional[str]:
    'Reinstall given application'
    pkgname, vdir = utils.get_package_from_arg(pkgname, args)
    if not vdir:
        return f'Application {pkgname} is not installed.'

    print(f'Reinstalling {pkgname} ..')
    pip_args = 'sync --compile --reinstall'.split() + \
            utils.make_args((args.verbose, '-v'))

    data = utils.get_json(vdir, args) or {}
    if url := data.get('url'):
        pip_args.extend(['-i', url])

    # Use explicit python (or reset) if given, else use what was
    # explicitly specified in original install, else use default
    # python
    nargs = copy(args)
    if args.reset_python:
        data.pop('python', None)
    elif args.python:
        data['python'] = args.python
    else:
        nargs.python = data.get('python')

    venv_args.extend(['-p', str(utils.get_python(nargs))])

    if args.system_site_packages:
        data['sys'] = True
    elif args.no_system_site_packages:
        data.pop('sys', None)

    if data.get('sys'):
        venv_args.append('--system-site-packages')

    with tempfile.TemporaryDirectory() as tdir:
        tfile = Path(tdir, args._freeze_file)
        shutil.copyfile(vdir / args._freeze_file, tfile)

        # Recreate the vdir
        if not run(venv_args + [str(vdir)]):
            utils.rm_vdir(vdir, args)
            return f'Error: failed to recreate {vdir} for {pkgname}.'

        if not utils.piprun(vdir, args, pip_args + [str(tfile)]):
            utils.rm_vdir(vdir, args)
            return f'Error: failed to resync {pkgname}'

    if err := utils.make_links(vdir, pkgname, args, data):
        return err

    print(f'{pkgname} reinstalled.')
    return None

def init(parser: ArgumentParser) -> None:
    'Called to add command arguments to parser at init'
    xgroup = parser.add_mutually_exclusive_group()
    xgroup.add_argument('-p', '--python',
                        help='specify explicit python executable path')
    xgroup.add_argument('--reset-python', action='store_true',
                        help='reset any explicit python path to default python')
    ygroup = parser.add_mutually_exclusive_group()
    ygroup.add_argument('--system-site-packages', action='store_true',
                        help='allow venv access to system packages, '
                        'overrides the per-application setting')
    ygroup.add_argument('--no-system-site-packages', action='store_true',
                        help='remove venv access to system packages, '
                        'overrides the per-application setting')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='give more output')
    parser.add_argument('--all', action='store_true',
                        help='reinstall ALL applications')
    parser.add_argument('--skip', action='store_true',
                        help='skip the specified applications when '
                        'reinstalling all (only can be specified with --all)')
    parser.add_argument('package', nargs='*',
                        help='application[s] to reinstall (or to skip for '
                        '--all --skip)')

def main(args: Namespace) -> Optional[str]:
    'Called to action this command'
    venv_args = [args._uv, 'venv'] + utils.make_args((args.verbose, '-v'),
                                                     (not args.verbose, '-q'))
    for pkgname in utils.get_package_names(args):
        if error := _reinstall(args, pkgname, venv_args.copy()):
            return error

    return None
